from typing import Any

from suds import sudsobject

from oda_wd_client.base.api import WorkdayClient
from oda_wd_client.base.utils import get_id_from_list
from oda_wd_client.service.financial_management.types import (
    Company,
    CostCenterWorktag,
    Currency,
    SpendCategory,
)
from oda_wd_client.service.resource_management.exceptions import NoSupplierID
from oda_wd_client.service.resource_management.types import (
    PrepaidAmortizationType,
    Supplier,
    SupplierInvoice,
    SupplierInvoiceLine,
    SupplierStatus,
    TaxApplicability,
    TaxCode,
    TaxOption,
    TaxRate,
    TaxRateOptionsData,
    TaxRecoverability,
)


def _get_account_data_from_dict(data: dict) -> dict:
    """
    Retrieve bank account data from a Settlement Account Widget Data dict

    The keys in the returned dict are defined by the Supplier pydantic model
    """
    settlement_data = data.get("Settlement_Account_Data", None)
    # TODO: Add filtering to ensure we use the correct account
    used = settlement_data[0] if settlement_data else {}

    ret = {
        "iban": used.get("IBAN", None),
        "bank_account": used.get("Bank_Account_Number", None),
    }

    return ret


def _get_contact_data_from_dict(data: dict) -> dict:
    """
    Retrieve contact information from a Contact Data dict

    The keys in the returned dict are defined by the Supplier pydantic model
    """
    ret = {}
    address_info = data.get("Address_Data", None)
    phone_info = data.get("Phone_Data", None)
    email_info = data.get("Email_Address_Data", None)
    url_info = data.get("Web_Address_Data", None)

    if address_info:
        current = sorted(
            address_info, reverse=True, key=lambda addr: addr["_Effective_Date"]
        )[0]
        # We'll just use the pre-formatted address from Workday as value
        ret["address"] = current["_Formatted_Address"].replace("&#xa;", "\n")

    if phone_info:
        # TODO: Decide if we need to filter on which number to use
        ret["phone"] = phone_info[0]["_E164_Formatted_Phone"]

    if email_info:
        # TODO: Decide if we need to filter on which email address to use
        ret["email"] = email_info[0]["Email_Address"]

    if url_info:
        # TODO: Decide if we need to filter on which URL to use
        ret["url"] = url_info[0]["Web_Address"]

    return ret


def workday_supplier_to_pydantic(data: dict) -> Supplier:
    """
    Parse a suds dict representing a supplier from Workday and return a Supplier pydantic instance
    """
    sup_data = data["Supplier_Data"]
    sup_id = sup_data.get("Supplier_ID", None)
    if not sup_id:
        raise NoSupplierID()
    primary_tax_id = None
    tax_id_data = sup_data.get("Tax_ID_Widget_Data", {}).get("Tax_ID_Data", [])
    for item in tax_id_data:
        if item["Primary_Tax_ID"]:
            primary_tax_id = item["Tax_ID_Text"]
    account_data = _get_account_data_from_dict(
        sup_data.get("Settlement_Account_Widget_Data", {})
    )
    contact_data = _get_contact_data_from_dict(
        sup_data["Business_Entity_Data"].get("Contact_Data", {})
    )
    currency_ref = sup_data.get("Currency_Reference", None)
    status_ref = None
    for status_line in sup_data.get("Supplier_Status_Data", []):
        _ref = SupplierStatus.from_id_list(status_line["Status_Reference"]["ID"])
        if _ref:
            status_ref = _ref

    return Supplier(
        workday_id=sup_id,
        reference_id=sup_data.get("Supplier_Reference_ID", None),
        name=sup_data["Supplier_Name"],
        status=status_ref,
        payment_terms=get_id_from_list(
            sup_data.get("Payment_Terms_Reference", {}).get("ID", []),
            "Payment_Terms_ID",
        ),
        primary_tax_id=primary_tax_id,
        # Currency_ID _should_ be in accordance with ISO 4217
        currency=get_id_from_list(currency_ref["ID"], "Currency_ID")
        if currency_ref
        else None,
        **contact_data,
        **account_data,
    )


def _workday_invoice_line_to_pydantic(data: dict, order: int) -> SupplierInvoiceLine:
    cost_center = None
    # Worktags is a list of tags, each with their own list of IDs
    worktags = data["Worktags_Reference"]
    for tag in worktags:
        _cost_center = CostCenterWorktag.from_id_list(tag["ID"])
        if _cost_center:
            cost_center = _cost_center

    tax_rate_options_data = None
    try:
        tax_option = TaxOption.from_id_list(data["Tax_Option_1_Reference"]["ID"])
        tax_rate = TaxRate.from_id_list(data["Tax_Rate_1_Reference"]["ID"])
        tax_recoverability = TaxRecoverability.from_id_list(
            data["Tax_Recoverability_1_Reference"]["ID"]
        )
        # Need to do type narrowing here to avoid passing on nullable objects to TaxRateOptionsData
        assert tax_option
        assert tax_rate
        assert tax_recoverability
        tax_rate_options_data = TaxRateOptionsData(
            tax_option=tax_option,
            tax_rate=tax_rate,
            tax_recoverability=tax_recoverability,
        )
    except (KeyError, AssertionError):
        pass

    return SupplierInvoiceLine(
        order=order,
        description=data.get("Item_Description", ""),
        tax_rate_options_data=tax_rate_options_data,
        tax_applicability=TaxApplicability.from_id_list(
            data["Tax_Applicability_Reference"]["ID"]
        ),
        tax_code=TaxCode.from_id_list(data["Tax_Code_Reference"]["ID"]),
        spend_category=SpendCategory.from_id_list(
            data["Spend_Category_Reference"]["ID"]
        ),
        cost_center=cost_center,
        gross_amount=data["Extended_Amount"],
    )


def workday_supplier_invoice_to_pydantic(data: dict) -> SupplierInvoice:
    data_list = data["Supplier_Invoice_Data"]
    assert len(data_list) == 1, "Expecting only one invoice in this dataset"
    inv: dict[str, Any] = data_list[0]

    lines = []
    for i, line in enumerate(
        sorted(
            inv["Invoice_Line_Replacement_Data"], key=lambda _line: _line["Line_Order"]
        )
    ):
        lines.append(_workday_invoice_line_to_pydantic(line, i))

    company_ref = get_id_from_list(
        inv["Company_Reference"]["ID"], "Company_Reference_ID"
    )
    currency_ref = get_id_from_list(inv["Currency_Reference"]["ID"], "Currency_ID")
    supplier_ref = get_id_from_list(inv["Supplier_Reference"]["ID"], "Supplier_ID")
    prepaid_ref = get_id_from_list(
        inv["Prepayment_Release_Type_Reference"]["ID"], "Prepayment_Release_Type_ID"
    )

    # Type narrowing
    assert company_ref is not None
    assert currency_ref is not None
    assert supplier_ref is not None

    return SupplierInvoice(
        workday_id=get_id_from_list(
            data["Supplier_Invoice_Reference"]["ID"], "Supplier_Invoice_Reference_ID"
        ),
        invoice_number=inv["Invoice_Number"],
        company=Company(workday_id=company_ref),
        currency=Currency(currency_code=currency_ref),
        supplier=Supplier(workday_id=supplier_ref),
        invoice_date=inv["Invoice_Date"],
        due_date=inv["Due_Date_Override"],
        total_amount=inv["Control_Amount_Total"],
        tax_amount=inv["Tax_Amount"],
        prepaid=inv["Prepaid"],
        prepayment_release_type_reference=PrepaidAmortizationType(
            workday_id=getattr(PrepaidAmortizationType.WorkdayID, prepaid_ref)
        )
        if prepaid_ref
        else None,
        lines=lines,
    )


def _get_wd_invoice_lines_from_invoice(
    client, lines: list[SupplierInvoiceLine]
) -> list[sudsobject.Object]:
    returned_lines = []

    for line in lines:
        wd_line = client.factory("ns0:Supplier_Invoice_Line_Replacement_DataType")
        wd_line.Line_Order = line.order
        wd_line.Quantity = line.quantity
        wd_line.Item_Description = line.description
        # Extended amount is without VAT
        wd_line.Extended_Amount = line.net_amount
        if line.spend_category:
            wd_line.Spend_Category_Reference = line.spend_category.wd_object(client)

        # Tax options
        tax_opts = line.tax_rate_options_data
        if tax_opts:
            wd_tax = client.factory("ns0:Tax_Rate_Options_DataType")
            wd_tax.Tax_Rate_1_Reference = tax_opts.tax_rate.wd_object(client)
            wd_tax.Tax_Recoverability_1_Reference = (
                tax_opts.tax_recoverability.wd_object(client)
            )
            wd_tax.Tax_Option_1_Reference = tax_opts.tax_option.wd_object(client)
            wd_line.Tax_Rate_Options_Data = wd_tax

        # Tax code
        if line.tax_applicability:
            wd_line.Tax_Applicability_Reference = line.tax_applicability.wd_object(
                client
            )
        if line.tax_code:
            wd_line.Tax_Code_Reference = line.tax_code.wd_object(client)

        # Worktags
        if line.cost_center:
            wd_line.Worktags_Reference.append(line.cost_center.wd_object(client))

        returned_lines.append(wd_line)

    return returned_lines


def pydantic_supplier_invoice_to_workday(
    invoice: SupplierInvoice, client: WorkdayClient
) -> sudsobject.Object:
    """
    Generate the data that is needed for a for Supplier_Invoice_Data in a call to Submit_Supplier_Invoice
    """
    invoice_data = client.factory("ns0:Supplier_Invoice_DataType")

    invoice_data.Submit = invoice.submit
    invoice_data.Locked_in_Workday = invoice.locked_in_workday

    # We want to let workday set the invoice number, but keep the invoice's number for supplier's invoice number
    invoice_data.Invoice_Number = None
    invoice_data.Suppliers_Invoice_Number = invoice.invoice_number
    invoice_data.Company_Reference = invoice.company.wd_object(client)
    invoice_data.Currency_Reference = invoice.currency.wd_object(client)
    invoice_data.Supplier_Reference = invoice.supplier.wd_object(client)
    if invoice.tax_option:
        invoice_data.Default_Tax_Option_Reference = invoice.tax_option.wd_object(client)
    invoice_data.Invoice_Date = str(invoice.invoice_date)
    invoice_data.Due_Date_Override = str(invoice.due_date)
    invoice_data.Control_Amount_Total = invoice.total_amount
    # invoice_data.Tax_Amount = invoice.tax_amount
    # invoice_data.Attachment_Data = _get_wd_attachment_data_from_invoice(invoice)
    invoice_data.Invoice_Line_Replacement_Data = _get_wd_invoice_lines_from_invoice(
        client, invoice.lines
    )
    if invoice.additional_reference_number:
        if not invoice.additional_type_reference:
            raise Exception(
                "If additional_reference_number is set, then additional_type_reference must also be set"
            )
        invoice_data.Additional_Reference_Number = invoice.additional_reference_number
        invoice_data.Additional_Type_Reference = (
            invoice.additional_type_reference.wd_object(client)
        )

    invoice_data.Prepaid = invoice.prepaid
    invoice_data.Prepayment_Release_Type_Reference = (
        (invoice.prepayment_release_type_reference.wd_object(client))
        if invoice.prepayment_release_type_reference
        else None
    )
    if invoice.attachments:
        invoice_data.Attachment_Data = [
            attachment.wd_object(client) for attachment in invoice.attachments
        ]

    return invoice_data



from typing import Iterable

from suds import sudsobject

from oda_wd_client.base.api import WorkdayClient
from oda_wd_client.base.tools import suds_to_dict
from oda_wd_client.service.staffing.types import Document
from oda_wd_client.service.staffing.utils import workday_document_to_pydantic


class Staffing(WorkdayClient):
    service = "Staffing"

    def put_document(self, *args, **kwargs) -> sudsobject.Object:
        """
        Upload a new document to WD
        """
        method = "Put_Worker_Document"
        return self._request(method, *args, Add_Only=True, **kwargs)

    def get_documents(
        self, raw_objects=False
    ) -> Iterable[Document | sudsobject.Object]:
        """
        Get all documents (as we can't filter on user...)
        """
        method = "Get_Worker_Documents"

        _response_group = self.factory("ns0:Worker_Document_Response_GroupType")
        _response_group.Include_Worker_Document_Data = False
        extra_request_kwargs = {
            "Worker_Document_Response_Group": _response_group,
        }
        docs = self._get_paginated(
            method,
            "Worker_Documents",
            per_page=25,
            extra_request_kwargs=extra_request_kwargs,
        )
        for doc in docs:
            if raw_objects:
                yield doc
            else:
                yield workday_document_to_pydantic(suds_to_dict(doc))



from typing import Iterator

from suds import sudsobject

from oda_wd_client.base.api import WorkdayClient
from oda_wd_client.base.tools import suds_to_dict
from oda_wd_client.service.resource_management.exceptions import NoSupplierID
from oda_wd_client.service.resource_management.types import Supplier, SupplierInvoice
from oda_wd_client.service.resource_management.utils import (
    pydantic_supplier_invoice_to_workday,
    workday_supplier_invoice_to_pydantic,
    workday_supplier_to_pydantic,
)


class ResourceManagement(WorkdayClient):
    service = "Resource_Management"

    def get_suppliers(
        self, return_suds_object=False
    ) -> Iterator[sudsobject.Object | Supplier]:
        method = "Get_Suppliers"
        results = self._get_paginated(method, "Supplier")
        for supplier in results:
            try:
                yield supplier if return_suds_object else workday_supplier_to_pydantic(
                    suds_to_dict(supplier)
                )
            except NoSupplierID:
                # We'll skip suppliers without an ID
                pass

    def get_supplier_invoices(
        self, return_suds_object=False
    ) -> Iterator[sudsobject.Object | SupplierInvoice]:
        method = "Get_Supplier_Invoices"
        results = self._get_paginated(method, "Supplier_Invoice")
        for invoice in results:
            yield invoice if return_suds_object else workday_supplier_invoice_to_pydantic(
                suds_to_dict(invoice)
            )

    def submit_supplier_invoice(self, invoice: SupplierInvoice) -> sudsobject.Object:
        method = "Submit_Supplier_Invoice"
        request = self.factory("ns0:Submit_Supplier_Invoice_Request")
        request.Supplier_Invoice_Data = pydantic_supplier_invoice_to_workday(
            invoice, self
        )
        return self._request(
            method, Supplier_Invoice_Data=request.Supplier_Invoice_Data
        )

    def cancel_supplier_invoice(self, invoice_ref_id: str):
        """
        Cancel an invoice in Workday

        :param invoice_ref_id:
        :return:
        """
        method = "Cancel_Supplier_Invoice"
        ref_obj = self.factory("ns0:Supplier_InvoiceObjectType")
        id_obj = self.factory("ns0:Supplier_InvoiceObjectIDType")
        id_obj._type = "Supplier_Invoice_Reference_ID"
        id_obj.value = invoice_ref_id
        ref_obj.ID.append(id_obj)
        return self._request(method, Supplier_Invoice_Reference=ref_obj)



from suds import sudsobject

from oda_wd_client.base.api import WorkdayClient
from oda_wd_client.base.utils import get_id_from_list
from oda_wd_client.service.financial_management.types import (
    AccountingJournalData,
    Company,
    ConversionRate,
    ConversionRateType,
    CostCenterWorktag,
    Currency,
    JournalEntryLineData,
    ProjectWorktag,
    SpendCategory,
)
from oda_wd_client.service.resource_management.types import TaxApplicability


def workday_conversion_rate_to_pydantic(data: dict) -> ConversionRate:
    """
    Create a ConversionRate pydantic object from a suds datadict from Workday
    """
    workday_id = get_id_from_list(
        data["Currency_Conversion_Rate_Reference"]["ID"], "WID"
    )
    assert len(data["Currency_Conversion_Rate_Data"]) == 1, (
        "Code is written expecting that we only have one currency "
        "rate data per object, but that is not the case here"
    )
    sub_data = data["Currency_Conversion_Rate_Data"][0]
    from_ref = get_id_from_list(
        sub_data["From_Currency_Reference"]["ID"], "Currency_ID"
    )
    target_ref = get_id_from_list(
        sub_data["Target_Currency_Reference"]["ID"], "Currency_ID"
    )
    type_id = get_id_from_list(
        sub_data["Currency_Rate_Type_Reference"]["ID"], "Currency_Rate_Type_ID"
    )

    assert workday_id, "All currency conversion rates need a Workday ID"
    assert from_ref, "All currency conversion rates need a 'from currency'-reference"
    assert (
        target_ref
    ), "All currency conversion rates need a 'target currency'-reference"
    assert type_id, "All currency conversion rates need a 'rate type'-reference"

    return ConversionRate(
        workday_id=workday_id,
        from_currency_iso=from_ref,
        to_currency_iso=target_ref,
        rate=sub_data["Currency_Rate"],
        rate_type_id=ConversionRate.RateTypeID(type_id),
        effective_timestamp=sub_data["Effective_Timestamp"],
    )


def workday_conversion_rate_type_to_pydantic(data: dict) -> ConversionRateType:
    """
    Create a ConversionRateType pydantic object from a suds datadict from Workday
    """
    workday_id = get_id_from_list(data["Currency_Rate_Type_Reference"]["ID"], "WID")
    assert workday_id, "All currency conversion rate types need a Workday ID"
    assert len(data["Currency_Rate_Type_Data"]) == 1, (
        "Code is written to expect that we only have one currency "
        "rate type data per object, but that is not the case here"
    )
    sub_data = data["Currency_Rate_Type_Data"][0]
    return ConversionRateType(
        workday_id=workday_id,
        text_id=sub_data["Currency_Rate_Type_ID"],
        description=sub_data["Currency_Rate_Type_Description"],
        is_default=sub_data["Currency_Rate_Type_Default"],
    )


def workday_company_to_pydantic(data: dict) -> Company:
    # For some weird reason, the object that holds the data for a specific company is wrapped in a list
    assert (
        len(data["Company_Data"]) == 1
    ), "Company_Data for each company should be singular"
    cdata = data["Company_Data"][0]

    currency_code = get_id_from_list(
        cdata["Accounting_Data"]["Currency_Reference"]["ID"], "Currency_ID"
    )
    # We'll expect that each company only has to care about taxation in one jurisdiction, and use the first object
    if "Tax_Status_Data" in cdata:
        country_code = get_id_from_list(
            cdata["Tax_Status_Data"][0]["Country_Reference"]["ID"],
            "ISO_3166-1_Alpha-2_Code",
        )
    else:
        country_code = None

    return Company(
        workday_id=cdata["Organization_Data"]["ID"],
        name=cdata["Organization_Data"]["Organization_Name"],
        country_code=country_code,
        currency=Currency(currency_code=currency_code) if currency_code else None,
    )


def workday_currency_to_pydantic(data: dict) -> Currency:
    return Currency(
        currency_code=data["Currency_ID"],
        description=data["Currency_Description"],
        retired=data["Currency_Retired"],
    )


def workday_tax_applicability_to_pydantic(data: dict) -> TaxApplicability:
    data = data["Tax_Applicability_Data"]
    return TaxApplicability(
        workday_id=data["Tax_Applicability_ID"],
        code=data["Tax_Applicability_Code"],
        taxable=data["Taxable"],
    )


def workday_cost_center_to_pydantic(data: dict) -> CostCenterWorktag:
    cost_center_id = get_id_from_list(
        data["Cost_Center_Reference"]["ID"], "Cost_Center_Reference_ID"
    )
    assert cost_center_id, "No ID of cost center found, this is required"
    return CostCenterWorktag(
        workday_id=cost_center_id,
        name=data["Cost_Center_Data"]["Organization_Data"]["Organization_Name"],
        active=data["Cost_Center_Data"]["Organization_Data"]["Organization_Active"],
    )


def workday_spend_category_to_pydantic(data: dict) -> SpendCategory:
    cat_data = data["Resource_Category_Data"]
    ref_id = get_id_from_list(
        data["Resource_Category_Reference"]["ID"], "Spend_Category_ID"
    )
    resource_id = cat_data["Resource_Category_ID"]
    assert (
        ref_id == resource_id
    ), "The Resource_Category_ID and Spend_Category_ID are expected to be equal"
    usage_ids = []

    for item in cat_data.get("Spend_Category_Usage_Reference", []):
        val = get_id_from_list(item["ID"], "WID")
        if val:
            usage_ids.append(val)

    return SpendCategory(
        workday_id=ref_id,
        name=cat_data["Resource_Category_Name"],
        inactive=cat_data["Inactive"],
        usage_ids=usage_ids,
    )


def workday_project_to_pydantic(data: dict) -> ProjectWorktag:
    proj_data = data["Basic_Project_Data"]
    return ProjectWorktag(
        workday_id=proj_data["Project_ID"],
        name=proj_data["Project_Name"],
        inactive=proj_data["Inactive"],
    )


def pydantic_conversion_rate_to_workday(
    rate: ConversionRate, client: WorkdayClient
) -> sudsobject.Object:
    """
    Create a suds object from a Pydantic object. Used for creating/updating conversion rates in Workday
    """
    rate_data = client.factory("ns0:Currency_Conversion_Rate_DataType")
    from_currency_id = client.factory("ns0:CurrencyObjectIDType")
    target_currency_id = client.factory("ns0:CurrencyObjectIDType")
    rate_type_id = client.factory("ns0:Currency_Rate_TypeObjectIDType")

    # Populate objects
    rate_data.Effective_Timestamp = rate.effective_timestamp
    rate_data.Currency_Rate = rate.rate
    from_currency_id.value = rate.from_currency_iso
    target_currency_id.value = rate.to_currency_iso
    rate_data.Calculate_Inverse_Rate = True

    # Stuff everything into the final object
    from_currency_id._type = "Currency_ID"
    target_currency_id._type = "Currency_ID"
    from_currency = client.factory("ns0:CurrencyObjectType")
    target_currency = client.factory("ns0:CurrencyObjectType")
    from_currency.ID.append(from_currency_id)
    target_currency.ID.append(target_currency_id)
    rate_type_id._type = "Currency_Rate_Type_ID"
    rate_type_id.value = rate.rate_type_id.value
    rate_type = client.factory("ns0:Currency_Rate_TypeObjectType")
    rate_type.ID.append(rate_type_id)
    rate_data.From_Currency_Reference = from_currency
    rate_data.Target_Currency_Reference = target_currency
    rate_data.Currency_Rate_Type_Reference = rate_type

    return rate_data


def _pydantic_journal_entry_line_to_workday(
    journal_line: JournalEntryLineData, client: WorkdayClient
) -> sudsobject.Object:
    wd_journal_entry_line = client.factory("ns0:Journal_Entry_Line_DataType")

    wd_journal_entry_line.Ledger_Account_Reference = (
        journal_line.ledger_account.wd_object(client)
    )

    if journal_line.spend_category:
        spend_category = journal_line.spend_category.wd_object(client)
        wd_journal_entry_line.Worktags_Reference.append(spend_category)

    if journal_line.cost_center:
        cost_center = journal_line.cost_center.wd_object(client)
        wd_journal_entry_line.Worktags_Reference.append(cost_center)

    wd_journal_entry_line.Debit_Amount = journal_line.debit
    wd_journal_entry_line.Credit_Amount = journal_line.credit
    wd_journal_entry_line.Memo = journal_line.memo

    return wd_journal_entry_line


def pydantic_accounting_journal_to_workday(
    journal: AccountingJournalData, client: WorkdayClient
) -> sudsobject.Object:
    """
    Create a suds object from a Pydantic object. Used for submitting accounting journals to Workday
    """
    # create container object and add data
    wd_accounting_journal = client.factory("ns0:Accounting_Journal_DataType")
    wd_accounting_journal.Accounting_Date = journal.accounting_date
    wd_accounting_journal.Accounting_Journal_ID = journal.accounting_journal_id
    wd_accounting_journal.Journal_Source_Reference = journal.journal_source.wd_object(
        client
    )
    if journal.company.currency:
        wd_accounting_journal.Currency_Reference = journal.company.currency.wd_object(
            client
        )
    if journal.memo:
        wd_accounting_journal.Journal_Entry_Memo = journal.memo

    if journal.submit:
        wd_accounting_journal.Submit = journal.submit

    wd_accounting_journal.Company_Reference = journal.company.wd_object(client)
    wd_accounting_journal.Ledger_Type_Reference = journal.ledger_type.wd_object(client)
    for journal_entry_line_data in journal.journal_entry_line_data:
        wd_data = _pydantic_journal_entry_line_to_workday(
            journal_entry_line_data, client
        )
        wd_accounting_journal.Journal_Entry_Line_Replacement_Data.append(wd_data)

    return wd_accounting_journal


def get_business_process_parameters(
    auto_complete: bool, client: WorkdayClient
) -> sudsobject.Object:
    financials_business_process_parameters = client.factory(
        "ns0:Financials_Business_Process_ParametersType"
    )
    financials_business_process_parameters.Auto_Complete = auto_complete
    return financials_business_process_parameters



try:
    from structlog import getLogger

    _structlog = True
except ImportError:
    from logging import getLogger

    _structlog = False


_logger = getLogger()


def log(method: str, *args, **kwargs) -> None:
    """
    Drop kwargs if we don't have structlog

    :param method: Log method, i.e. "error" or "debug"
    :param args: positional args which will be passed on as-is to the logger
    :param kwargs: kwargs will be passed on to structlog and dropped else-wise
    :return: None
    """
    if _structlog:
        getattr(_logger, method)(*args, **kwargs)
    else:
        getattr(_logger, method)(*args)



from datetime import datetime

# As defined by API docs
WORKDAY_DATE_FORMAT = "%m/%d/%Y"


def get_id_from_list(id_list: list, id_type: str) -> str | None:
    """
    Workday operates with a lot of fields that contains list of IDs. We often want to extract a single value of
    these, based on the type of the ID.

    Will return the first ID matching the type (there should only be one), or None if no ID matches the type.

    Args:
        id_list: List of ID dicts (not suds objects) from Workday
        id_type: The value we'll use to filter on _type
    """
    ids = [ref["value"] for ref in id_list if ref["_type"] == id_type]
    if ids:
        return ids[0]
    return None


def parse_workday_date(val):
    if isinstance(val, str):
        return datetime.strptime(val, WORKDAY_DATE_FORMAT).date()
    return val



from enum import Enum

from pydantic import BaseModel

# All public imports should be done through oda_wd_client.types.staffing
__all__: list = []


class Document(BaseModel):
    """
    Reference: https://community.workday.com/sites/default/files/file-hosting/productionapi/Staffing/v40.2/Put_Worker_Document.html#Worker_Document_DataType  # noqa
    """

    class WorkdayCategory(str, Enum):
        employee_contract = "EMPLOYEE_CONTRACT"
        student_collections = "STUDENT_COLLECTIONS"
        prospect_resume_cover_letter = "PROSPECT_RESUME_AND_COVER_LETTER"
        candidate_resume_cover_letter = "CANDIDATE_RESUME_AND_COVER_LETTER"
        residency_determination = "RESIDENCY_DETERMINATION"
        education = "EDUCATION"
        multimedia_audio = "MULTI_MEDIA_AUDIO"
        multimedia_video = "MULTI_MEDIA_VIDEO"
        drive_document_template = "DRIVEDOCUMENTTEMPLATE"
        committees = "COMMITTEES"
        recommendation = "RECOMMENDATION"
        position = "POSITION"
        credential = "CREDENTIAL"
        retiree = "RETIREE"
        application = "APPLICATION"
        secondary = "SEC"
        action_item = "ACTION_ITEM"
        reference_letter = "REFERENCE_LETTER"
        accommodation_verification = "ACCOMMODATION_VERIFICATION"
        probation_period = "PROBATION_PERIOD"
        language_proficiency = "LANGUAGE_PROFICIENCY"
        time_off = "TIME_OFF"
        portfolio = "PORTFOLIO"
        resume = "RESUME"
        writing_sample = "WRITING_SAMPLE"
        benefits = "BENEFITS"
        verification = "VERIFICATION"
        certification = "CERT"
        company_policy_related = "COMPANY_POLICY_RELATED"
        leave_of_absence = "LOA"
        personal_information = "PERSONAL_INFORMATION"
        time_tracking = "TIME_TRACKING"
        offers = "OFFERS"
        background_check = "BACKGROUND CHECK"
        assessment = "ASSESSMENT"
        interview = "INTERVIEW"
        other_documents = "OTHER_DOCUMENTS"
        agency = "AGENCY"
        academic_appointments = "ACADEMIC_APPOINTMENTS"
        period_activity_pay = "PERIOD_ACTIVITY_PAY"
        named_professorships = "NAMED_PROFESSORSHIPS"
        notice_period = "NOTICE_PERIOD"
        licenses = "LICENSES"
        passports_and_visas = "PASSPORTS_AND_VISAS"
        dependents = "DEPENDENTS"
        employment_agreement = "EMPLOYMENT_AGREEMENT"
        other_student = "OTHER_STUDENT"
        international_student = "INTERNATIONAL_STUDENT"

    class OdaCategory(str, Enum):
        # TODO: Make this model extendable so that we do not have to hardcode our values here
        tax_forms = "DOC_CAT_Tax_Forms"
        non_disclosure_agreement = "DOC_CAT_Non-Disclosure_Agreement"
        id_verification = "DOC_CAT_ID_Verification"
        other_documents_offboarding = "DOC_CAT_Other_Documents_Offboarding"
        compensation = "DOC_CAT_Compensation"
        transfer = "DOC_CAT_Transfer"
        promotion = "DOC_CAT_Promotion"
        nav_documents = "DOC_CAT_NAV_Documents"
        education_reimbursement = "DOC_CAT_Education_Reimbursement"
        contingent_worker_contract = "DOC_CAT_Contingent_Worker_Contract"
        termination = "DOC_CAT_Termination"
        doctors_notice = "DOC_CAT_Doctors_notice"
        hire = "DOC_CAT_Hire"
        sick_leave = "DOC_CAT_Sick_Leave"
        internal_documents = "DOC_CAT_Internal_documents_(not_visible_for_employee)"
        other_employee_documents = "DOC_CAT_Other_employee_documents"

    employee_number: str
    wd_owned_category: WorkdayCategory | None
    oda_category: OdaCategory | None
    filename: str
    comment: str



from suds.sudsobject import Object, asdict  # type: ignore


def suds_to_dict(d: Object) -> dict:
    """
    Convert Suds object into serializable format.

    Source: https://gist.github.com/mattkatz/65bbc17dbad94c97a01a472734b65d50
    """
    out: dict[str | dict, str | dict | list] = {}
    for k, v in asdict(d).items():
        if hasattr(v, "__keylist__"):
            out[k] = suds_to_dict(v)
        elif isinstance(v, list):
            key_list = []
            for item in v:
                if hasattr(item, "__keylist__"):
                    key_list.append(suds_to_dict(item))
                else:
                    key_list.append(item)
            out[k] = key_list
        else:
            out[k] = v
    return out



import os
from datetime import datetime
from typing import Any, Iterator
from urllib.parse import urljoin

from suds import client as suds_client
from suds import sudsobject, wsse
from suds.plugin import MessagePlugin

from oda_wd_client.base.logging import log


class SudsHax(MessagePlugin):
    def marshalled(self, context):
        # Ugly hack: Add namespace to ID attribute (ref.
        # https://stackoverflow.com/a/5730868/6866905)
        root = context.envelope.getChild("Body")

        def recurse(obj):
            for attr in obj.attributes:
                attr.prefix = obj.parent.prefix
            if obj.children:
                for child in obj.children:
                    recurse(child)

        recurse(root)


class SudsLog(MessagePlugin):
    """
    Write all responses and requests to files during development

    Inspired by https://www.guguweb.com/2019/12/11/how-to-log-raw-xml-requests-and-responses-in-suds/
    """

    def __init__(self, location):
        self.location = location
        if not os.path.exists(location):
            os.makedirs(location)

    def pretty_log(self, xml_content, direction):
        try:
            file_path = os.path.join(
                self.location, f"suds-log-{datetime.now()}-{direction}.xml"
            )
            with open(file_path, "wb") as f:
                f.write(xml_content)
        except Exception as e:
            log("warning", f"Cannot log XML ({e})")

    def sending(self, context):
        self.pretty_log(context.envelope, "send")

    def received(self, context):
        self.pretty_log(context.reply, "receive")


# Using a singleton to hold all of our client instances to avoid re-initiation during app lifecycle. Changing the
# credentials requires us to reload the app anyway.
_workday_clients: dict[str, suds_client.Client] = {}


class WorkdayClient:
    """
    An abstraction on top of the Suds SOAP client to give us an interface into
    Workday that's been instantiated with our own settings

    Relevant docs: https://community.workday.com/articles/628676
    """

    # Each SOAP service must be instantiated individually, and we need more than one...
    _services: list[str] = [
        "Human_Resources",
        "Staffing",
        "Financial_Management",
        "Resource_Management",
    ]

    service: str

    def __init__(
        self,
        base_url: str,
        tenant_name: str,
        username: str,
        password: str,
        lazy_init: bool = True,
        log_location: str | None = None,
    ) -> None:
        self._auth_base_url = base_url
        self._auth_tenant_name = tenant_name
        self._auth_username = username
        self._auth_password = password
        self._log_location = log_location
        # If we do lazy init, we only initialize each service when they're first needed, since there's an overhead
        # with the init process
        if not lazy_init:
            for service in self._services:
                self._init_service_client(service)
        assert (
            self.service
        ), "`WorkdayClient` must be subclassed, and all subclasses must have `.service` set"

    def _init_service_client(self, service: str) -> None:
        if service not in _workday_clients:
            url = self._get_client_url(service)
            _workday_clients[service] = self._setup_client(url)

    def _get_client_url(self, service: str) -> str:
        return urljoin(
            self._auth_base_url,
            f"/ccx/service/{self._auth_tenant_name}/{service}/v38.1?wsdl",
        )

    def _setup_client(self, url: str) -> suds_client.Client:
        """
        Create client instance with secrets and config
        """
        plugins = [SudsHax()]
        if self._log_location:
            plugins.append(SudsLog(self._log_location))
        client = suds_client.Client(url, plugins=plugins)
        security = wsse.Security()
        token = wsse.UsernameToken(
            f"{self._auth_username}@{self._auth_tenant_name}",
            self._auth_password,
        )
        token.setnonce()
        token.setcreated()
        security.tokens.append(token)
        client.set_options(wsse=security, prettyxml=True)
        return client

    def get_client(self, service: str) -> suds_client.Client:
        """
        Instantiate client if it hasn't been instantiated, else just get client instance
        """
        if service not in _workday_clients:
            self._init_service_client(service)
        return _workday_clients[service]

    def _request(self, method_name: str, *args, **kwargs) -> sudsobject.Object:
        """
        Wrapper to call service methods on client instance
        """
        method = getattr(self.get_client(self.service).service, method_name)
        return method(*args, **kwargs)

    def _get_paginated(
        self,
        method_name: str,
        results_key: str,
        filters: dict[str, Any] | None = None,
        per_page: int = 100,
        extra_request_kwargs: dict[str, Any] | None = None,
    ) -> Iterator[sudsobject.Object]:
        extra_request_kwargs = extra_request_kwargs or {}
        page = 1
        max_page = None
        while max_page is None or page <= max_page:
            str_max_page = "Unknown"
            if max_page:
                str_max_page = int(max_page)
            log(
                "info",
                f"Getting page {page} of {str_max_page}",
            )
            _filter = self.factory("ns0:Response_FilterType")

            if filters:
                for key, val in filters.items():
                    setattr(_filter, key, val)

            _filter.Page = page
            _filter.Count = per_page
            response = self._request(
                method_name, Response_Filter=_filter, **extra_request_kwargs
            )

            # Response_Results / Response_Data might be wrapped in a list, if so, pick out the object
            # Double check that they only have one item
            response_results = response.Response_Results

            if isinstance(response_results, list):
                assert (
                    len(response_results) == 1
                ), "If Response_Results is a list, it should only have one item"
                response_results = response_results[0]

            # If we get no results, we won't have any response data attribute, so we need to do an early return
            #  before attempting to access that data.
            if int(response_results.Total_Results) == 0:
                log("info", "No results")
                return

            response_data = response.Response_Data

            if isinstance(response_data, list):
                assert (
                    len(response_data) == 1
                ), "If Response_Data is a list, it should only have one item"
                response_data = response_data[0]

            max_page = response_results.Total_Pages
            for result in getattr(response_data, results_key):
                yield result

            page += 1

    def factory(self, name: str) -> sudsobject.Object:
        """
        Convenience wrapper
        """
        return self.get_client(self.service).factory.create(name)



"""
Convenience imports to make all types available through one path. Implementing applications should not import from
oda_wd_client.service, which holds the business logic, but rather use this point of entry instead.

Usage:
    from oda_wd_client.types.$SERVICE import $TYPE

Example:
    from oda_wd_client.types.human_resources import Worker
"""
